﻿using System;
using System.Collections.Generic;
using System.Text;
using Stunnware.Tools.CrmService3;
using System.Xml;
using System.Web.Services.Protocols;
using System.IO;

namespace Stunnware.Crm.Tools {

   /// <summary>
   /// The CrmLogService wraps a CRM service to get to the SOAP messages being sent and received.
   /// </summary>
   class CrmLogService : CrmService {

      /// <summary>
      /// Subscribe to this event to get notified when a SOAP message is sent or received.
      /// </summary>
      public event SoapMessageEventHandler SoapMessageSerializationFinished;

      /// <summary>
      /// Standard event handling method.
      /// </summary>
      /// <param name="e"></param>
      protected void OnSoapMessageSerializationFinished(SoapMessageEventArgs e) {
         if (SoapMessageSerializationFinished != null) {
            SoapMessageSerializationFinished(this, e);
         }
      }

      /// <summary>
      /// GetReaderForMessage is called when a SOAP document was received from the CRM server.
      /// </summary>
      /// <param name="message">The received message</param>
      /// <param name="bufferSize">The buffer size with which to initialize the XmlTextReader.</param>
      /// <returns>An XmlReader used to parse the SOAP message</returns>
      protected override XmlReader GetReaderForMessage(SoapClientMessage message, int bufferSize) {

         //Read the entire message
         StreamReader reader = new StreamReader(message.Stream);
         string soapClientMessageXml = reader.ReadToEnd();

         //Raise the SoapMessageSerializationFinished event
         SoapMessageEventArgs e = new SoapMessageEventArgs(message, SoapMessageType.Response, soapClientMessageXml);
         OnSoapMessageSerializationFinished(e);

         //As the message.Stream is read-only and does not support positioning, we copy the entire content
         //to a new stream and use this for further processing.
         MemoryStream ms = new MemoryStream();
         byte[] data = Encoding.UTF8.GetBytes(soapClientMessageXml);
         ms.Write(data, 0, data.Length);
         ms.Position = 0;

         XmlTextReader xmlTextReader = new XmlTextReader(ms);
         return xmlTextReader;
      }

      /// <summary>
      /// GetWriterForMessage is called when a SOAP message is to be sent to the CRM server.
      /// </summary>
      /// <param name="message">The message to send</param>
      /// <param name="bufferSize">The buffer size with which to initialize the XmlTextWriter.</param>
      /// <returns></returns>
      protected override XmlWriter GetWriterForMessage(SoapClientMessage message, int bufferSize) {

         //We're using our own XmlWriter implementation to get to the xml used in the SOAP message
         XmlLogWriter writer = new XmlLogWriter(message, bufferSize);

         //If you have subscribed to the SoapMessageSerializationFinished event, the subsrciption is
         //propagated to the XmlLogWriter. 
         if (SoapMessageSerializationFinished != null) {
            writer.SoapMessageSerializationFinished += new SoapMessageEventHandler(writer_SoapMessageSerializationFinished);
         }

         return writer;
      }

      /// <summary>
      /// Receives notifications from our XmlLogWriter and passes it to the subscribers of the
      /// CrmLogService.SoapMessageSerializationFinished event.
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="e"></param>
      void writer_SoapMessageSerializationFinished(object sender, SoapMessageEventArgs e) {
         OnSoapMessageSerializationFinished(e);
      }
   }
}
