﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;
using System.Web.Services.Protocols;

namespace Stunnware.Crm.Tools {

   /// <summary>
   /// The XmlLogWriter is an implementation of the abstract XmlWriter class. It uses an XmlTextWriter
   /// as a helper class. The purpose of this class is to get to the SOAP message stream.
   /// </summary>
   class XmlLogWriter : XmlWriter {

      /// <summary>
      /// Subscribe to this event to get notified when a SOAP message is sent or received.
      /// </summary>
      public event SoapMessageEventHandler SoapMessageSerializationFinished;

      SoapClientMessage _soapClientMessage;
      XmlTextWriter _xmlTextWriter;
      MemoryStream _memoryStream;
      long _bytesCopied;

      /// <summary>
      /// Creates a new instance of this class.
      /// </summary>
      /// <param name="message"></param>
      /// <param name="bufferSize"></param>
      public XmlLogWriter(SoapClientMessage message, int bufferSize) {
         _soapClientMessage = message;
         _memoryStream = new MemoryStream();
         _xmlTextWriter = new XmlTextWriter(_memoryStream, Encoding.UTF8);
      }

      /// <summary>
      /// Standard event handling.
      /// </summary>
      /// <param name="messageType"></param>
      /// <param name="soapMessageXml"></param>
      protected void OnSoapMessageSerializationFinished(SoapMessageType messageType, string soapMessageXml) {
         if (SoapMessageSerializationFinished != null) {
            SoapMessageEventArgs e = new SoapMessageEventArgs(_soapClientMessage, messageType, soapMessageXml);
            SoapMessageSerializationFinished(this, e);
         }
      }

      /// <summary>
      /// Returns the SOAP message in its XML format. If we still need more data, the XmlDocument.LoadXml
      /// will fail and an exception is thrown. This is the only reason why the XmlDocument is used.
      /// Whenever doc.LoadXml() succeeds, we have the entire SOAP message and can raise the 
      /// SoapMessageSerializationFinished event.
      /// </summary>
      /// <returns></returns>
      private string GetSoapMessageXml() {

         long currentMemoryStreamPosition = _memoryStream.Position;
         _memoryStream.Position = 0;

         string xml;

         StreamReader reader = new StreamReader(_memoryStream, Encoding.UTF8);
         xml = reader.ReadToEnd();

         XmlDocument doc = new XmlDocument();
         doc.LoadXml(xml);

         return xml;
      }

      /// <summary>
      /// Whenever the Flush method is invoked, we must copy the buffered data to the original
      /// SOAP stream. We also test if the entire message has been written and raise the 
      /// SoapMessageSerializationFinished event when this has happened.
      /// </summary>
      public override void Flush() {
         _xmlTextWriter.Flush();

         long bytesToCopy = _memoryStream.Position - _bytesCopied;

         if (bytesToCopy > 0) {
            _memoryStream.Seek(_bytesCopied, SeekOrigin.Begin);

            byte[] newData = new byte[bytesToCopy];

            _memoryStream.Read(newData, 0, (int) bytesToCopy);
            _soapClientMessage.Stream.Write(newData, 0, (int) bytesToCopy);
            _soapClientMessage.Stream.Flush();

            _bytesCopied += bytesToCopy;

            try {
               string soapMessageXml = GetSoapMessageXml();
               OnSoapMessageSerializationFinished(SoapMessageType.Request, soapMessageXml);
            }

            catch {
            }
         }
      }

      public override void Close() {
         _xmlTextWriter.Close();
      }

      public override string LookupPrefix(string ns) {
         return _xmlTextWriter.LookupPrefix(ns);
      }

      public override void WriteBase64(byte[] buffer, int index, int count) {
         _xmlTextWriter.WriteBase64(buffer, index, count);
      }

      public override void WriteCData(string text) {
         _xmlTextWriter.WriteCData(text);
      }

      public override void WriteCharEntity(char ch) {
         _xmlTextWriter.WriteCharEntity(ch);
      }

      public override void WriteChars(char[] buffer, int index, int count) {
         _xmlTextWriter.WriteChars(buffer, index, count);
      }

      public override void WriteComment(string text) {
         _xmlTextWriter.WriteComment(text);
      }

      public override void WriteDocType(string name, string pubid, string sysid, string subset) {
         _xmlTextWriter.WriteDocType(name, pubid, sysid, subset);
      }

      public override void WriteEndAttribute() {
         _xmlTextWriter.WriteEndAttribute();
      }

      public override void WriteEndDocument() {
         _xmlTextWriter.WriteEndDocument();
      }

      public override void WriteEndElement() {
         _xmlTextWriter.WriteEndElement();
      }

      public override void WriteEntityRef(string name) {
         _xmlTextWriter.WriteEntityRef(name);
      }

      public override void WriteFullEndElement() {
         _xmlTextWriter.WriteFullEndElement();
      }

      public override void WriteProcessingInstruction(string name, string text) {
         _xmlTextWriter.WriteProcessingInstruction(name, text);
      }

      public override void WriteRaw(string data) {
         _xmlTextWriter.WriteRaw(data);
      }

      public override void WriteRaw(char[] buffer, int index, int count) {
         _xmlTextWriter.WriteRaw(buffer, index, count);
      }

      public override void WriteStartAttribute(string prefix, string localName, string ns) {
         _xmlTextWriter.WriteStartAttribute(prefix, localName, ns);
      }

      public override void WriteStartDocument(bool standalone) {
         _xmlTextWriter.WriteStartDocument(standalone);
      }

      public override void WriteStartDocument() {
         _xmlTextWriter.WriteStartDocument();
      }

      public override void WriteStartElement(string prefix, string localName, string ns) {
         _xmlTextWriter.WriteStartElement(prefix, localName, ns);
      }

      public override WriteState WriteState {
         get { return _xmlTextWriter.WriteState; }
      }

      public override void WriteString(string text) {
         _xmlTextWriter.WriteString(text);
      }

      public override void WriteSurrogateCharEntity(char lowChar, char highChar) {
         _xmlTextWriter.WriteSurrogateCharEntity(lowChar, highChar);
      }

      public override void WriteWhitespace(string ws) {
         _xmlTextWriter.WriteWhitespace(ws);
      }
   }

}
